/***
 *
 * Copyright (C) 2003 International Business Machines, Inc.
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * 	need api in libsysfs to get one attribute of one particular device.
 *	also an api to get all attributes of one device.
 *
 */

#include "sysdiag.h"


int sysdiag_db_open(char *dbfile, int option, int mode)
{
	sysdiag_db = tdb_open(dbfile, 0, option, mode, 0644);
	if (sysdiag_db == NULL)
		return -1;
	return 0;

}

int sysdiag_db_close()
{
	if (tdb_close(sysdiag_db) != 0) {
		fprintf(stdout, "error closing the database!!\n");
		return -1;
	}
	return 0;

}

/**
 * this will get the number of elements seprated by # are there.
 *
 */
int sysdiag_db_num_of_elements(char *data)
{
	char *temp = NULL;
	int num = 0;
	
	if (data == NULL)
		return 0;
	
	temp = calloc(1, strlen(data));
	if (temp == NULL) {
		perror("calloc failed");
		return 0;
	}
		
	memcpy(temp, data, strlen(data));
	strtok(temp, "#");
	num++;
	
	while (strtok(NULL, "#") != NULL) {
		num++;
	}
	free(temp);
	return num;
}

/**
 * 
 */
char * sysdiag_db_get_str(char *data, int entry)
{
	char *temp = NULL;
	char *s = NULL;
	char *str = NULL;
	int num = 0;

	if (data == NULL)
		return NULL;
	temp = calloc(1, strlen(data) +1);
	if (temp == NULL) {
		perror("calloc failed\n");
		return NULL;
	}
	strncpy(temp, data, strlen(data));
	s = strtok(temp, "#");
	num++;
	while( num < entry) {
		s = strtok(NULL, "#");
		if (s == NULL)
			break;
		num++;
	}

	if (s == NULL) {
		free(temp);
		return NULL;
	}
	str = calloc(1, strlen(s)+1);
	if (str == NULL) {
		perror( "calloc failed");
		free(temp);
		return NULL;
	}
	strncpy(str, s, strlen(s));
	free(temp);
	return (char *)str;

}

/**
 * this will search for the element in the data and then return the element
 * number.
 * if element not found then it returns 0;
 */
int sysdiag_db_find_element(char *data, char *element)
{
	int num = 0, i;
	char *s = NULL;
	
	
	if ((num = sysdiag_db_num_of_elements(data)) == 0) {
		fprintf(stdout, "no of elements  = %d\n", num);
		return 0;
	}	
	for (i = 1; i <= num; i++) {
		if ((s = sysdiag_db_get_str(data, i)) == NULL) {
			return 0;
		}
		if (strcmp(element, s) == 0) {
			free(s);
			return i;
		}
	}	
	return 0;
}



/**
 * this will fetch a record from the DB.
 *
 */
char * sysdiag_get_record(char *keystr)
{
	TDB_DATA data, key;
	char *datastr = NULL;

	if (keystr == NULL)
		return NULL;

	key.dptr = keystr;
	key.dsize = strlen(keystr) +1;

	data = tdb_fetch(sysdiag_db, key);
	if (data.dptr == NULL || data.dsize == 0) 
		return NULL;
	datastr = data.dptr;
	return datastr;
}

/**
 * this will add a record in the database.
 */
int sysdiag_add_record(char *keystr, char *datastr)
{
	TDB_DATA data, key;

	if ( geteuid() != 0) {
		fprintf(stdout, "database open in read only mode, " 
				"cannot add records\n");
		return -1;
	}

	if ( keystr == NULL || datastr == NULL) {
		return -1;
	}
	data.dptr = datastr;
	data.dsize = strlen(datastr)+1;
	key.dptr = keystr;
	key.dsize = strlen(keystr)+1;

	 return tdb_store(sysdiag_db, key, data, TDB_REPLACE);

}

/**
 * this will first get the list of block device there are. and then
 * for each device that is there, then open the block device and put
 * the physical device information in the db.
 *
 */
int load_block_dev()
{
	char *data = NULL, *str = NULL;
	char keystr[SYSFS_NAME_LEN];
	char datastr[SYSDIAG_DB_DATA_SIZE];
	struct sysfs_class *class = NULL;
	struct dlist *list = NULL;
	struct sysfs_class_device *clsdev = NULL;
	struct sysfs_device *dev = NULL;
	
	if ((data = sysdiag_get_record(SYSDIAG_CLASS_STR)) == NULL)
		return -1;
	if (strstr(data, SYSFS_BLOCK_NAME) == NULL) {
		return -1;		
	}
	if ((class = sysfs_open_class(SYSFS_BLOCK_NAME)) == NULL )
		return -1;

	if (class->devices == NULL) {
		return -1;
	}
	list = class->devices;
	dlist_for_each_data(list, clsdev, struct sysfs_class_device) {
		if (clsdev->sysdevice == NULL) {
			continue;
		}
		dev = clsdev->sysdevice;
		
		memset(keystr, 0, SYSFS_NAME_LEN);
		memset(datastr, 0, SYSDIAG_DB_DATA_SIZE);
		strncpy(keystr, SYSFS_BLOCK_NAME, strlen(SYSFS_BLOCK_NAME));
		strncat(keystr, "#", 1);
		strncat(keystr, clsdev->name, strlen(clsdev->name));
		strncpy(datastr, dev->bus_id, strlen(dev->bus_id));
		
		sysdiag_add_record(keystr, datastr);
						
	}
	sysfs_close_class(class);
	return 0;
}

/**
 * this will load the class device information.
 */
int load_class()
{
	struct sysfs_class *class = NULL;
	struct sysfs_class_device *dev = NULL;
	struct dlist *list = NULL;
	char *data = NULL, *str = NULL;
	char temp[SYSFS_NAME_LEN], key[SYSFS_NAME_LEN];
	char *datastr = NULL;
	int num = 0, i = 0, j=0;
	struct sysfs_device *child = NULL;
	
	data = sysdiag_get_record(SYSDIAG_CLASS_STR);
	if (data == NULL) {
		fprintf(stdout, "fialed getting bus information from db\n");
		return -1;
		
	}
	if ((num = sysdiag_db_num_of_elements(data)) == 0) {
		fprintf(stdout, "no data availble in the database for Bus\n");
		return -1;
	}
	for (i = 1; i <= num ; i++) {
		if ((str = sysdiag_db_get_str(data, i)) == NULL) {
			break;
		}
		if ((class = sysfs_open_class(str)) == NULL) {
			fprintf(stdout, "failed opening bus: %s\n", str);
			continue;
		}
		memset(key, 0, SYSFS_NAME_LEN);
		strncpy(key, SYSDIAG_CLASS_STR, strlen(SYSDIAG_CLASS_STR));
		strncat(key, "#", 1);
		strncat(key, SYSFS_DEVICES_NAME, strlen(SYSFS_DEVICES_NAME));
		strncat(key, "#", 1);
		strncat(key, class->name, strlen(class->name));
		
		if (class->devices == NULL) {
			sysfs_close_class(class);
			continue;
		}
		j = 0;
		dlist_for_each_data(class->devices, dev, 
				struct sysfs_class_device) {
			j += strlen(dev->name);
			j++;
		}
		datastr = calloc(1, j);
		j = 0;
		dlist_for_each_data(class->devices, dev, 
				struct sysfs_class_device) {
			j++;
			if ( j > 1)
				strncat(datastr, "#", 1);
			strncat(datastr, dev->name, strlen(dev->name));
		}
		sysdiag_add_record(key, datastr);
		free(datastr);
		datastr = NULL;
		sysfs_close_class(class);
	}
	return 0;

}


/**
 * this will load the bus device information in the data base.
 */
int load_bus()
{
	struct sysfs_bus *bus = NULL;
	struct sysfs_device *dev = NULL;
	struct sysfs_driver *drv = NULL;
	struct dlist *list = NULL;
	char *data = NULL, *str = NULL;
	char temp[SYSFS_NAME_LEN], key[SYSFS_NAME_LEN];
	char *datastr = NULL;
	int num = 0, i = 0, j=0;
	
	data = sysdiag_get_record(SYSDIAG_BUS_STR);
	if (data == NULL) {
		fprintf(stdout, "fialed getting bus information from db\n");
		return -1;
		
	}
	if ((num = sysdiag_db_num_of_elements(data)) == 0) {
		fprintf(stdout, "no data availble in the database for Bus\n");
		return -1;
	}
	for (i = 1; i <= num ; i++) {
		if ((str = sysdiag_db_get_str(data, i)) == NULL) {
			break;
		}
		if ((bus = sysfs_open_bus(str)) == NULL) {
			fprintf(stdout, "failed opening bus: %s\n", str);
			continue;
		}
		memset(key, 0, SYSFS_NAME_LEN);
		strncpy(key, SYSDIAG_BUS_STR, strlen(SYSDIAG_BUS_STR));
		strncat(key, "#", 1);
		strncat(key, SYSFS_DEVICES_NAME, strlen(SYSFS_DEVICES_NAME));
		strncat(key, "#", 1);
		strncat(key, bus->name, strlen(bus->name));
		
		if (bus->devices == NULL) {
			sysfs_close_bus(bus);
			continue;
		}
		j = 0;
		dlist_for_each_data(bus->devices, dev, struct sysfs_device) {
			j += strlen(dev->bus_id);
			j++;
		}
		datastr = calloc(1, j);
		j = 0;
		dlist_for_each_data(bus->devices, dev, struct sysfs_device) {
			j++;
			if ( j > 1)
				strncat(datastr, "#", 1);
			strncat(datastr, dev->bus_id, strlen(dev->bus_id));
		}
		sysdiag_add_record(key, datastr);

		free(datastr);
		datastr = NULL;

		memset(key, 0, SYSFS_NAME_LEN);
		strncpy(key, SYSDIAG_BUS_STR, strlen(SYSDIAG_BUS_STR));
		strncat(key, "#", 1);
		strncat(key, SYSFS_DRIVERS_NAME, strlen(SYSFS_DRIVERS_NAME));
		strncat(key, "#", 1);
		strncat(key, bus->name, strlen(bus->name));
		
		if (bus->drivers == NULL) {
			sysfs_close_bus(bus);
			continue;
		}
		j = 0;
		dlist_for_each_data(bus->drivers, drv, struct sysfs_driver) {
			j += strlen(drv->name);
			j++;
		}
		datastr = calloc(1, j);
		j = 0;
		dlist_for_each_data(bus->drivers, drv, struct sysfs_driver) {
			j++;
			if ( j > 1)
				strncat(datastr, "#", 1);
			strncat(datastr, drv->name, strlen(drv->name));
		}
		sysdiag_add_record(key, datastr);
		
		sysfs_close_bus(bus);
	}
	return 0;
}


/**
 * this will load the default information in the db.
 */
int load_default(char *subsys, char *key)
{
	struct dlist *list = NULL;
	char *cur = NULL;
	char data[SYSDIAG_DB_DATA_SIZE];
	
	memset(data, 0, SYSDIAG_DB_DATA_SIZE);
	
	if ((list = sysfs_open_subsystem_list(subsys)) != NULL) {
		dlist_for_each_data(list, cur, char) {
			if ((strlen(cur) + strlen(data) + 1) > 
					SYSDIAG_DB_DATA_SIZE)
				break;
			if (strlen(data) > 0)
				strncat(data, "#", 1);
			strncat(data, cur, strlen(cur));
		}
		sysfs_close_list(list);
		list = NULL;
	}

	return sysdiag_add_record(key, data);
	
}


/**
 * this will load the basic sysdiag information in the sysdiag DB.
 */
int load_sysdiag_db()
{

	if (load_default(SYSFS_BUS_DIR, SYSDIAG_BUS_STR) != 0) {
		fprintf(stderr, "loading bus information failed !!\n");
		return -1;
	}
	if (load_default(SYSFS_CLASS_DIR, SYSDIAG_CLASS_STR) != 0) {
		fprintf(stderr, "loading class information failed !!\n");
		return -1;
	}
	
	if (load_default(SYSFS_DEVICES_DIR, SYSFS_DEVICES_NAME) != 0) {
		fprintf(stderr, "loading root device information failed !!\n");
		return -1;
	}

	load_block_dev();
	load_bus();
	load_class();
	
}


/**
 * this will setup the whole sysdiag db. this will check if the db already
 * exists. If it does it will delete it and reload it.
 */
int sysdiag_db_main(void)
{
	int option =0;
	int mode= 0;
	int flag = 0;
	struct stat buf;
	
	option = TDB_DEFAULT;

	if (geteuid() !=0 )
		mode = O_RDONLY;
	else
		mode = O_RDWR | O_CREAT;
		
	
	if (DIAG_OPTIONS & DIAG_DB_RELOAD)
		option |= TDB_CLEAR_IF_FIRST;

	if (stat(sysdiag_db_file, &buf) == -1) {
		flag = 1;
		
	}
	
	if (sysdiag_db_open((char *)sysdiag_db_file, option, mode ) == -1) {
		fprintf(stderr, "sysdiag_db_main: error opening database :%s\n"
				, sysdiag_db_file);
		return -1;
	}
	if (mode == O_RDONLY && flag == 0) {
		return 0;
	}
	if (load_sysdiag_db() == -1) {
		fprintf(stderr, "sysdiag_db_main: error loading database :%s\n"
				, sysdiag_db);
		return -1;
	}
	return 0;

}
